from pathlib import Path

from loguru import logger

from ..data.basic import read_json

PYMEMRI_FOLDER = ".pymemri"
POD_KEYS_FOLDER = "pod_keys"
POD_KEYS_FILENAME = "keys.json"
POD_KEYS_FULL_FOLDER = Path.home() / ".pymemri" / POD_KEYS_FOLDER
DEFAULT_POD_KEY_PATH = POD_KEYS_FULL_FOLDER / POD_KEYS_FILENAME


def read_pod_key(key_type, file=DEFAULT_POD_KEY_PATH):
    try:
        json = read_json(file)
    except:
        raise ValueError(
            f"Trying to read key from {file}, but file or key does not exist"
        ) from None
    try:
        key = json[key_type]
        logger.info(f"reading {key_type} from {file}")
        return key
    except:
        raise ValueError(f"{key_type} not specified in {file}") from None



"""
    flask.helpers
    ~~~~~~~~~~~~~

    Implements various helpers.

    :copyright: 2010 Pallets
    :license: BSD-3-Clause
"""
import os
import pkgutil
import sys
from threading import Lock


def get_root_path(import_name):
    """
    NOTE: This code is a copy-paste from Flask
    Returns the path to a package or cwd if that cannot be found.  This
    returns the path of a package or the folder that contains a module.

    Not to be confused with the package path returned by :func:`find_package`.
    """
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, "__file__"):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == "__main__":
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, "get_filename"):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, "__file__", None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError(
                "No root path can be found for the provided "
                'module "%s".  This can happen because the '
                "module came from an import hook that does "
                "not provide file name information or because "
                "it's a namespace package.  In this case "
                "the root path needs to be explicitly "
                "provided." % import_name
            )

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


class DictProxy:
    def __init__(self):
        self._dict = dict()
        self._lock = Lock()

    def get(self, key, default=None):
        with self._lock:
            return self._dict.get(key, default)

    def set(self, key, value):
        with self._lock:
            self._dict[key] = value



import abc
import importlib
import inspect
import json
import warnings
from abc import ABCMeta

from loguru import logger

from pymemri import PodClient
from pymemri.data.basic import read_json
from pymemri.webserver.public_api import ENDPOINT_METADATA, register_endpoint

from ..data.basic import write_json
from ..data.schema import Account, PluginRun
from ..webserver.webserver import WebServer
from .authenticators.credentials import PLUGIN_DIR
from .listeners import get_abort_plugin_listener, get_pod_restart_listener
from .states import RUN_COMPLETED, RUN_DAEMON, RUN_INITIALIZED, RUN_STARTED


class PluginBase(metaclass=ABCMeta):
    """Base class for plugins"""

    schema_classes = []

    def __init__(self, pluginRun: PluginRun = None, *, client: PodClient, **kwargs):
        super().__init__()
        if pluginRun is None:
            warnings.warn(
                "Plugin needs a pluginRun as kwarg, running without will only work in development.",
                RuntimeWarning,
            )
        self.pluginRun = pluginRun

        if client is None:
            raise ValueError("Plugins need a `client: PodClient` as kwarg to run.")
        self.client = client
        self._status_listeners = []
        self._config_dict = kwargs
        self._daemon = False

        if pluginRun is None:
            self._webserver = WebServer(8080)
        else:
            self._webserver = WebServer(pluginRun.webserverPort or 8080)

        self.endpoints = self._get_endpoint_methods()
        self._register_api_endpoints()

        self.set_run_status(RUN_INITIALIZED)

    def set_run_status(self, status):
        # TODO sync before setting status (requires pod_client.sync())
        if self.pluginRun and self.client:
            self.pluginRun.status = status
            self.client.update_item(self.pluginRun)

        self._status = status

    def set_progress(self, progress):
        if self.pluginRun and self.client:
            self.pluginRun.progress = progress
            self.client.update_item(self.pluginRun)

    def setup(self):
        if self.client and self.pluginRun:
            status_abort_listener = get_abort_plugin_listener(self.client, self.pluginRun.id)
            pod_restart_listener = get_pod_restart_listener(self.client, self.pluginRun.id)
            self._status_listeners.extend([status_abort_listener, pod_restart_listener])

        self._webserver.run()

    def teardown(self):
        for listener in self._status_listeners:
            listener.stop()

        self._webserver.shutdown()

    def _run(self):
        self.set_run_status(RUN_STARTED)

        self.setup()
        self.run()

        if self.daemon:
            self.set_run_status(RUN_DAEMON)
        else:
            self.teardown()
            self.set_run_status(RUN_COMPLETED)

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError()

    def add_to_schema(self):
        """
        Add all schema classes required by the plugin to self.client here.
        """
        if len(self.schema_classes):
            self.client.add_to_schema(*self.schema_classes)

    @classmethod
    def get_schema(cls):
        schema = []
        for schema_cls in cls.schema_classes:
            schema.extend(schema_cls.pod_schema())
        return schema

    def _register_api_endpoints(self):
        """Collect decorated methods and add them to the webserver routes"""
        for (path, method), endpoint in self.endpoints.items():
            self._webserver.app.add_api_route(path=path, endpoint=endpoint, methods=[method])

    def _get_endpoint_methods(self):
        """Collect decorated methods, bind them with `self`, and store in `endpoints`"""
        endpoints = {}
        for _method_name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(method, ENDPOINT_METADATA):
                metadata = method.__endpoint_metadata__

                if metadata in endpoints:
                    raise RuntimeError(
                        f"endpoint {metadata[0]} with method {metadata[1]} is already registered"
                    )

                endpoints[metadata] = method

        return endpoints

    @register_endpoint("/v1/health", "GET")
    def health_endpoint(self):
        """Returns current state of the plugin"""
        return self._status

    @register_endpoint("/v1/api", "GET")
    def get_public_api(self):
        """Returns exposed functions of the plugin"""

        def get_friendly_annotation_name(kv):
            """Annotation can be done by class, like: int, list, str
            or by alias from typing module, like List, Sequence, Tuple.
            """
            k, v = kv
            if hasattr(v, "__name__"):
                # For classes, use __name__ that returns
                # more concise value
                return (k, v.__name__)
            else:
                # The typing aliases do not provide __name__ attribute, so use
                # __repr__ implementation.
                return (k, str(v))

        resp = {
            func_name: {
                "method": method,
                "args": dict(
                    map(
                        get_friendly_annotation_name,
                        inspect.getfullargspec(func).annotations.items(),
                    )
                ),
            }
            for ((func_name, method), func) in self.endpoints.items()
        }
        return resp

    @property
    def daemon(self) -> bool:
        return self._daemon

    @daemon.setter
    def daemon(self, daemon: bool):
        """Setting to True will not close the plugin after calling run(), default if False"""
        self._daemon = daemon


class PluginError(Exception):
    """Generic class for plugin errors. This error is raised when a plugin raises an unexpected exception."""

    pass


def write_run_info(plugin, id_):
    try:
        if plugin is None:
            raise ValueError("Empty container")
        run_path = PLUGIN_DIR / plugin / "current_run.json"
        run_path.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"writing run info to {run_path}")
        write_json({"id": id_}, run_path)
    except Exception as e:
        logger.error(f"""failed to write run info to {run_path}\n{e}""")


def get_plugin_cls(plugin_module, plugin_name):
    try:
        module = importlib.import_module(plugin_module)
        plugin_cls = getattr(module, plugin_name)
        return plugin_cls
    except (ImportError, AttributeError):
        raise ImportError(f"Unknown plugin: {plugin_module}.{plugin_name}")


def run_plugin_from_run_id(run_id, client, **kwargs):
    """
    Runs a plugin from run_id, initialized with **kwargs.

    Args:
        client (PodClient): client containing PluginRun
        run_id (int): id of the PluginRun
    """

    run = client.get(run_id)
    write_run_info(
        run.pluginModule.split(".")[0] if run.pluginModule is not None else run.containerImage,
        run.id,
    )

    plugin_cls = get_plugin_cls(run.pluginModule, run.pluginName)
    plugin = plugin_cls(pluginRun=run, client=client, **kwargs)
    plugin.add_to_schema()

    plugin._run()

    return plugin


def parse_metadata(fn, remove_container=False):
    metadata = read_json(fn)
    for k in ["pluginModule", "pluginName"]:
        if k not in metadata:
            raise ValueError(f"Missing metadata: {k}")

    run_vars = {k: v for k, v in metadata.items() if k in PluginRun.properties}
    run = PluginRun.from_json(run_vars)
    if remove_container:
        run.containerImage = "none"

    if "account" in metadata:
        account = Account.from_json(metadata["account"])
        run.add_edge("account", account)
    return run


def parse_config(run_config, config_file=None, remove_container=False):
    """
    Parse the configuration of the plugin. A configuration is a dict that is passed to the plugin init as kwargs.
    If configuration file is defined, the run_config is ignored.
    """
    if config_file is not None:
        config = read_json(config_file)
    elif isinstance(run_config, str) and len(run_config):
        config = json.loads(run_config)
    else:
        config = dict()

    if not isinstance(config, dict):
        raise ValueError(f"Incorrect plugin config format, expected a dict, got a {type(config)}")
    return config


def create_run_expanded(client, run):
    client.create(run)
    accounts = run.account
    if accounts:
        account = accounts[0]
        client.create(account)
        client.create_edge(run.get_edges("account")[0])



from typing import TYPE_CHECKING, Any, List

if TYPE_CHECKING:
    from ..data.schema import Item
    from ..pod.client import PodClient


class Query:
    def __init__(self, *properties: List[str]):
        """
        A Query implements functionality to retrieve data from the pod to a tabular format.

        Given a list of `properties`, the `execute` method queries the pod for a set of given items,
        and retrieves the properties for each item if it exists. Note that a properties can be nested behind
        multiple edges, such as "sender.owner.firstName".
        """
        self.properties = list(properties)

    def traverse_edges(
        self, client: "PodClient", items: List["Item"], edges: List[str]
    ) -> List["Item"]:
        items = items.copy()

        for edge in edges:
            ids_to_query = list()
            query_item_idx = list()
            for i in range(len(items)):
                item = items[i]
                if item is None:
                    continue
                # Replace item with target item. If the edge is empty, it has to be queried again.
                try:
                    if edge not in item.edges:
                        items[i] = None
                    else:
                        items[i] = getattr(item, edge)[0]
                except Exception:
                    ids_to_query.append(item.id)
                    query_item_idx.append(i)
                    items[i] = None

            new_items = client.search({"ids": ids_to_query})
            for i, new_item in zip(query_item_idx, new_items):
                try:
                    items[i] = getattr(new_item, edge)[0]
                except Exception:
                    items[i] = None
        return items

    def get_property_values(self, client: "PodClient", prop: str, items: List["Item"]) -> list:
        edges, prop_name = self.parse_property(prop)
        target_items = self.traverse_edges(client, items, edges)

        result = [getattr(item, prop_name, None) for item in target_items]
        return result

    @staticmethod
    def parse_property(prop: str):
        prop = prop.split(".")
        edges = prop[:-1]
        prop = prop[-1]
        return edges, prop

    def convert_dtype(self, result, dtype):
        if dtype == "dict":
            return result
        elif dtype == "list":
            return [result[prop] for prop in self.properties]
        elif dtype in {"pandas", "pd", "df"}:
            try:
                import pandas as pd
            except ImportError:
                raise ImportError(
                    "Pandas is not installed." "Please install pandas to use this feature."
                )
            return pd.DataFrame.from_dict(result)
        else:
            raise ValueError(f"Unknown dtype: {dtype}")

    def execute(self, client: "PodClient", items: List["Item"], dtype="dict") -> Any:
        result = {prop: self.get_property_values(client, prop, items) for prop in self.properties}
        return self.convert_dtype(result, dtype)



import json
import zipfile
from pathlib import Path

Path.ls = lambda x: list(x.iterdir())
PYI_HOME = Path.cwd().parent
PYI_TESTDATA = PYI_HOME / "test" / "data"
HOME_DIR = Path.home()
MODEL_DIR = HOME_DIR / ".memri" / "models"
MEMRI_S3 = "https://memri.s3-eu-west-1.amazonaws.com"
MODEL_DIR.mkdir(parents=True, exist_ok=True)


def read_json(path):
    with open(path) as json_file:
        return json.load(json_file)


def write_json(obj, fname, indent=4):
    with open(fname, "w") as file_out:
        json.dump(obj, file_out, indent=indent)


def unzip(f, dest):
    with zipfile.ZipFile(str(f)) as zf:
        zf.extractall(str(dest))



import abc
from time import sleep

from ..states import RUN_FAILED, RUN_USER_ACTION_COMPLETED, RUN_USER_ACTION_NEEDED


class OAuthAuthenticator(metaclass=abc.ABCMeta):

    SLEEP_INTERVAL = 1.0

    def __init__(self, client, pluginRun):
        self.client = client
        self.pluginRun = pluginRun

    def authenticate(self):
        account = self.pluginRun.account[0] if len(self.pluginRun.account) > 0 else None
        if account.accessToken:
            if self.verify_access_token(account.accessToken):
                return

        tokens = None
        try:
            if not self.pluginRun.account[0].refreshToken:
                raise Exception("Refresh token is empty")
            tokens = self.refresh_tokens(self.pluginRun.account[0].refreshToken)
        except:  # no account exists or expired refresh token
            url = self.get_oauth_url()
            self.present_url_to_user(url)
            code = self.poll_for_code()
            tokens = self.get_tokens_from_code(code)
            # if not tokens: raise an exception or set pluginRun.status=FAILED and pluginRun.message=ERROR_MESSAGE

        self.store_tokens(tokens)

    def poll_for_code(self):
        while True:
            sleep(self.SLEEP_INTERVAL)
            self.pluginRun = self.client.get(self.pluginRun.id)
            if self.pluginRun.status == RUN_USER_ACTION_COMPLETED:
                return self.pluginRun.account[0].code
            if self.pluginRun.status == RUN_FAILED:
                raise Exception(f"Error in plugin.authenticators.oauth {self.pluginRun.message}")

    def present_url_to_user(self, url):
        # request user to visit url
        self.pluginRun.authUrl = url
        self.pluginRun.status = RUN_USER_ACTION_NEEDED
        self.pluginRun.update(self.client)

    def store_tokens(self, tokens):
        account = self.pluginRun.account[0]
        account.accessToken = tokens["access_token"]
        account.refreshToken = tokens["refresh_token"] if "refresh_token" in tokens else None
        account.update(self.client)

    def get_account(self):
        return self.pluginRun.account[0]

    @abc.abstractmethod
    def get_oauth_url(self):
        raise NotImplementedError

    @abc.abstractmethod
    def get_tokens_from_code(self, code):
        """Gets access and refresh tokens from 3rd party service
        and returns them in form:
            {
                'access_token': '...',
                'refresh_token': '...'
            }
        """
        raise NotImplementedError

    @abc.abstractmethod
    def refresh_tokens(self):
        """Gets new tokens by using an existing refresh token
        and returns them in form:
            {
                'access_token': '...',
                'refresh_token': '...'
            }

        """
        # use self.pluginRun.account[0].refreshToken
        raise NotImplementedError

    @abc.abstractmethod
    def verify_access_token(self, token):
        """
        Check if existing token is working. If this returns True, then user interaction will not be needed.
        """
        raise NotImplementedError



import getpass
import time
from time import sleep

from fastscript import Param, call_parse
from loguru import logger

from ...cvu.utils import get_default_cvu
from ...pod.client import DEFAULT_POD_ADDRESS, PodClient
from ...pod.utils import read_pod_key
from ..states import RUN_USER_ACTION_COMPLETED, RUN_USER_ACTION_NEEDED
from .credentials import PLUGIN_DIR, read_json, read_username_password


class PasswordAuthenticator:
    DEFAULT_CVU = "password_auth.cvu"
    MAX_LOGIN_ATTEMPTS = 3
    SLEEP_INTERVAL = 1.0
    MAX_POLL_TIME = 600

    def __init__(self, client, pluginRun):
        self.client = client
        self.pluginRun = pluginRun
        self.isTest = False

    def authenticate(self, login_callback):
        self.request_user_credentials()

        login_success = False
        for i in range(self.MAX_LOGIN_ATTEMPTS):
            username, password = self.poll_credentials()
            try:
                login_callback(username, password)
                login_success = True
                break
            except Exception as e:
                logger.error("Login failed, invalid credentials.")
                if self.pluginRun.account:
                    attempts_remaining = self.MAX_LOGIN_ATTEMPTS - (i + 1)
                    account = self.pluginRun.account[0]
                    account.errorMessage = (
                        f"Reached max login attempts. {attempts_remaining} attempts remaining"
                    )
                    self.client.update_item(account)

        if not login_success:
            self.pluginRun.status = "error"
            self.client.update_item(self.pluginRun)

            if self.pluginRun.account:
                account = self.pluginRun.account[0]
                account.errorMessage = "Reached max login attempts."
                self.client.update_item(account)

            raise RuntimeError("Reached max login attempts.")

    def request_user_credentials(self):
        cvu = get_default_cvu(self.DEFAULT_CVU)
        self.client.create(cvu)
        self.pluginRun.add_edge("view", cvu)
        self.client.create_edge(self.pluginRun.get_edges("view")[0])
        self.pluginRun.status = RUN_USER_ACTION_NEEDED
        self.client.update_item(self.pluginRun)

    def poll_credentials(self):
        start_time = time.time()
        while True:
            if time.time() - start_time > self.MAX_POLL_TIME:
                self.pluginRun.status = "error"
                self.client.update_item(self.pluginRun)
                raise RuntimeError("Stop polling, max time reached.")

            sleep(self.SLEEP_INTERVAL)
            self.pluginRun = self.client.get(self.pluginRun.id)
            logger.info(
                f"polling for credentials from account of pluginRun {self.pluginRun.id} ... run.status={self.pluginRun.status}",
            )
            if self.pluginRun.status == RUN_USER_ACTION_COMPLETED:
                account = self.pluginRun.account[0]
                return account.identifier, account.secret


def set_account_credentials(pod_client, run_id, username, password):
    run = pod_client.get(run_id)
    account = run.account[0]
    account.identifier = username
    account.secret = password
    run.status = "ready"

    pod_client.update_item(account)
    pod_client.update_item(run)
    print(f"Setting username and password for {run_id}'s Account'")


@call_parse
def simulate_enter_credentials(
    run_id: Param("run id that requires the password", str) = None,
    plugin: Param("plugin name, used for finding credentials", str) = None,
    pod_full_address: Param("The pod full address", str) = DEFAULT_POD_ADDRESS,
    database_key: Param("Database key of the pod", str) = None,
    owner_key: Param("Owner key of the pod", str) = None,
):
    if database_key is None:
        database_key = read_pod_key("database_key")
    if owner_key is None:
        owner_key = read_pod_key("owner_key")
    client = PodClient(url=pod_full_address, database_key=database_key, owner_key=owner_key)
    if plugin is not None:
        try:
            username, password = read_username_password(plugin)
        except Exception as e:
            logger.error(
                f"Could not find credentials for plugin {plugin}, if you don't want to set those locally remove the --plugin arg. Exiting"
            )

            exit()
    else:
        username = input(f"Enter username for service used by {run_id}: ")
        password = getpass.getpass(f"Enter password for service used by {run_id}: ")

    if run_id is None:
        try:
            run_path = PLUGIN_DIR / plugin / "current_run.json"
            logger.info(f"Reading run id from {run_path}")
            run_id = read_json(run_path)["id"]
        except Exception as e:
            logger.error("No run_id specified and could not read current run information")
    set_account_credentials(client, run_id, username, password)



# AUTOGENERATED, DO NOT EDIT!
# This file was generated by /tools/generate_central_schema.sh
# Visit https://gitlab.memri.io/memri/schema to learn more

from datetime import datetime
from typing import List, Optional, Union

from pymemri.data.schema.itembase import ItemBase


class Item(ItemBase):
    # Properties
    id: Optional[str] = None
    dateCreated: Optional[datetime] = None
    dateModified: Optional[datetime] = None
    dateServerModified: Optional[datetime] = None
    deleted: Optional[bool] = None
    externalId: Optional[str] = None
    isMock: Optional[bool] = None

    # Edges
    language: List["Language"] = []
    translation: List["Translation"] = []
    label: List["CategoricalPrediction"] = []


class Account(Item):
    # Properties
    avatarUrl: Optional[str] = None
    authEmail: Optional[str] = None
    code: Optional[str] = None
    displayName: Optional[str] = None
    externalId: Optional[str] = None
    handle: Optional[str] = None
    identifier: Optional[str] = None
    isMe: Optional[bool] = None
    itemType: Optional[str] = None
    secret: Optional[str] = None
    service: Optional[str] = None
    accessToken: Optional[str] = None
    refreshToken: Optional[str] = None
    description: Optional[str] = None

    # Edges
    changelog: List["AuditItem"] = []
    cryptoTransaction: List["CryptoTransaction"] = []
    location: List["Location"] = []
    network: List["Network"] = []
    ownCurrency: List["CryptoCurrency"] = []
    owner: List["Person"] = []
    trust: List["Account"] = []
    profilePicture: List["Photo"] = []
    following: List["Account"] = []
    follower: List["Account"] = []


class AuditItem(Item):
    # Properties
    actionname: Optional[str] = None
    content: Optional[str] = None
    date: Optional[datetime] = None


class CVUStoredDefinition(Item):
    # Properties
    definition: Optional[str] = None
    domain: Optional[str] = None
    itemType: Optional[str] = None
    name: Optional[str] = None
    queryStr: Optional[str] = None
    renderer: Optional[str] = None
    selector: Optional[str] = None
    definitionType: Optional[str] = None


class CategoricalPrediction(Item):
    # Properties
    source: Optional[str] = None
    value: Optional[str] = None
    probs: Optional[str] = None

    # Edges
    model: List["Model"] = []


class CreativeWork(Item):
    # Properties
    abstract: Optional[str] = None
    content: Optional[str] = None
    datePublished: Optional[datetime] = None
    itemType: Optional[str] = None
    keyword: Optional[str] = None
    textContent: Optional[str] = None
    title: Optional[str] = None
    transcript: Optional[str] = None

    # Edges
    contentLocation: List["Location"] = []
    file: List["File"] = []
    locationCreated: List["Location"] = []
    writtenBy: List["Person"] = []


class CryptoCurrency(Item):
    # Properties
    myToken: Optional[float] = None
    name: Optional[str] = None
    topic: Optional[str] = None

    # Edges
    currencySetting: List["CurrencySetting"] = []
    picture: List["Photo"] = []


class CryptoKey(Item):
    # Properties
    active: Optional[bool] = None
    itemType: Optional[str] = None
    keystr: Optional[str] = None
    name: Optional[str] = None
    role: Optional[str] = None
    starred: Optional[bool] = None

    # Edges
    owner: List["Person"] = []


class CryptoTransaction(Item):
    # Properties
    outward: Optional[bool] = None
    quantity: Optional[float] = None

    # Edges
    cryptoCurrency: List["CryptoCurrency"] = []
    relateToOther: List["Account"] = []
    relateToOwner: List["Account"] = []


class CurrencySetting(Item):
    # Properties
    deviceAddress: Optional[str] = None
    profileAddress: Optional[str] = None
    seedPhrase: Optional[str] = None
    tokenAddress: Optional[str] = None

    # Edges
    wallet: List["Wallet"] = []


class Dataset(Item):
    # Properties
    name: Optional[str] = None
    queryStr: Optional[str] = None

    # Edges
    entry: List["DatasetEntry"] = []
    feature: List["ItemPropertySchema"] = []
    labellingTask: List["LabellingTask"] = []
    datasetType: List["DatasetType"] = []


class DatasetEntry(Item):
    # Properties
    skippedByLabeller: Optional[bool] = None

    # Edges
    data: List[Union["Message", "Tweet", "EmailMessage"]] = []
    annotation: List["CategoricalLabel"] = []


class DatasetType(Item):
    # Properties
    name: Optional[str] = None
    queryStr: Optional[str] = None

    # Edges
    datasetPlugin: List["Plugin"] = []


class Diet(Item):
    # Properties
    abstract: Optional[str] = None
    content: Optional[str] = None
    datePublished: Optional[datetime] = None
    duration: Optional[int] = None
    itemType: Optional[str] = None
    keyword: Optional[str] = None
    name: Optional[str] = None
    textContent: Optional[str] = None
    title: Optional[str] = None
    transcript: Optional[str] = None

    # Edges
    contentLocation: List["Location"] = []
    file: List["File"] = []
    locationCreated: List["Location"] = []
    writtenBy: List["Person"] = []


class File(Item):
    # Properties
    filename: Optional[str] = None
    keystr: Optional[str] = None
    nonce: Optional[str] = None
    sha256: Optional[str] = None
    starred: Optional[bool] = None
    externalId: Optional[str] = None

    def __hash__(self):
        return hash(self.sha256)


class Integrator(Item):
    # Properties
    name: Optional[str] = None
    repository: Optional[str] = None


class ItemEdgeSchema(Item):
    # Properties
    edgeName: Optional[str] = None
    sourceType: Optional[str] = None
    targetType: Optional[str] = None


class ItemPropertySchema(Item):
    # Properties
    itemType: Optional[str] = None
    propertyName: Optional[str] = None
    valueType: Optional[str] = None


class Label(Item):
    # Properties
    color: Optional[str] = None
    name: Optional[str] = None


class LabelAnnotation(Item):
    # Properties
    allowSharing: Optional[bool] = None
    isSubmitted: Optional[bool] = None


class LabelOption(Item):
    # Properties
    color: Optional[str] = None
    name: Optional[str] = None


class LabellingDataType(Item):
    # Properties
    name: Optional[str] = None
    labelType: Optional[str] = None


class LabellingTask(Item):
    # Properties
    name: Optional[str] = None
    currentLabelOption: Optional[str] = None

    # Edges
    taskType: List["TextClassification"] = []
    labelOption: List["LabelOption"] = []
    view: List["CVUStoredDefinition"] = []


class Language(Item):
    # Properties
    languageCode: Optional[str] = None
    languageName: Optional[str] = None


class Location(Item):
    # Properties
    latitude: Optional[float] = None
    longitude: Optional[float] = None


class MediaObject(Item):
    # Properties
    bitrate: Optional[int] = None
    duration: Optional[int] = None
    endTime: Optional[datetime] = None
    fileLocation: Optional[str] = None
    startTime: Optional[datetime] = None

    # Edges
    file: List["File"] = []
    includes: List["Person"] = []


class MedicalCondition(Item):
    # Properties
    conditiontype: Optional[str] = None
    itemType: Optional[str] = None
    name: Optional[str] = None


class MessageChannel(Item):
    # Properties
    encrypted: Optional[bool] = None
    externalId: Optional[str] = None
    name: Optional[str] = None
    topic: Optional[str] = None
    service: Optional[str] = None
    isMock: Optional[bool] = None
    sourceProject: Optional[str] = None

    # Edges
    photo: List["Photo"] = []
    receiver: List["Account"] = []


class Model(Item):
    # Properties
    name: Optional[str] = None
    version: Optional[str] = None


class NavigationItem(Item):
    # Properties
    icon: Optional[str] = None
    itemType: Optional[str] = None
    sequence: Optional[int] = None
    sessionName: Optional[str] = None
    title: Optional[str] = None


class Network(Item):
    # Properties
    name: Optional[str] = None

    # Edges
    website: List["Website"] = []


class OauthFlow(Item):
    # Properties
    accessToken: Optional[str] = None
    accessTokenSecret: Optional[str] = None
    refreshToken: Optional[str] = None
    service: Optional[str] = None


class Person(Item):
    # Properties
    addressBookId: Optional[str] = None
    birthDate: Optional[datetime] = None
    deathDate: Optional[datetime] = None
    displayName: Optional[str] = None
    email: Optional[str] = None
    firstName: Optional[str] = None
    gender: Optional[str] = None
    lastName: Optional[str] = None
    role: Optional[str] = None
    sexualOrientation: Optional[str] = None
    starred: Optional[bool] = None

    # Edges
    account: List["Account"] = []
    address: List["Address"] = []
    birthPlace: List["Location"] = []
    cryptoKey: List["CryptoKey"] = []
    deathPlace: List["Location"] = []
    diet: List["Diet"] = []
    hasPhoneNumber: List["PhoneNumber"] = []
    label: List["Label"] = []
    me: List["Person"] = []
    medicalCondition: List["MedicalCondition"] = []
    mergedFrom: List["Person"] = []
    profilePicture: List["Photo"] = []
    relationship: List["Relationship"] = []
    website: List["Website"] = []


class PhoneNumber(Item):
    # Properties
    phoneNumber: Optional[str] = None


class Plugin(Item):
    # Properties
    bundleImage: Optional[str] = None
    containerImage: Optional[str] = None
    configJson: Optional[str] = None
    config: Optional[str] = None
    dataType: Optional[str] = None
    icon: Optional[str] = None
    pluginDescription: Optional[str] = None
    name: Optional[str] = None
    pluginModule: Optional[str] = None
    pluginName: Optional[str] = None
    pluginType: Optional[str] = None
    gitProjectId: Optional[int] = None

    # Edges
    view: List["CVUStoredDefinition"] = []
    templateSettings: List["TemplateSettings"] = []
    project: List["Project"] = []
    run: List["PluginRun"] = []


class PluginRun(Item):
    # Properties
    authUrl: Optional[str] = None
    containerImage: Optional[str] = None
    config: Optional[str] = None
    pluginModule: Optional[str] = None
    pluginName: Optional[str] = None
    progress: Optional[float] = None
    status: Optional[str] = None
    targetItemId: Optional[str] = None
    error: Optional[str] = None
    containerId: Optional[str] = None
    webserverPort: Optional[int] = None

    # Edges
    plugin: List["Plugin"] = []
    view: List["CVUStoredDefinition"] = []
    account: List["Account"] = []
    trigger: List["Trigger"] = []


class Post(Item):
    # Properties
    externalId: Optional[str] = None
    message: Optional[str] = None
    postDate: Optional[datetime] = None
    postType: Optional[str] = None
    isMock: Optional[bool] = None
    sourceProject: Optional[str] = None

    # Edges
    author: List["Account"] = []
    comment: List["Post"] = []
    parent: List["Post"] = []
    photo: List["Photo"] = []


class Project(Item):
    # Properties
    name: Optional[str] = None
    gitlabUrl: Optional[str] = None
    dataSource: Optional[str] = None

    # Edges
    dataset: List["Dataset"] = []
    labellingPlugin: List["Plugin"] = []


class Relationship(Item):
    # Properties
    label: Optional[str] = None
    proximityValue: Optional[int] = None

    # Edges
    relationship: List["Person"] = []


class SuggestedMerge(Item):
    # Properties
    score: Optional[float] = None
    task: Optional[str] = None

    # Edges
    mergeFrom: List["Person"] = []


class TemplateSettings(Item):
    # Properties
    templateName: Optional[str] = None
    templateId: Optional[int] = None
    dataSource: Optional[str] = None

    # Edges
    labelOption: List["LabelOption"] = []


class Trigger(Item):
    # Properties
    action: Optional[str] = None
    filterCreatedAfter: Optional[datetime] = None
    filterCreatedAfterPropertyName: Optional[str] = None
    pluginRunId: Optional[str] = None
    triggerOn: Optional[str] = None

    # Edges
    trigger: List["PluginRun"] = []


class VoteAction(Item):
    # Properties
    dateExecuted: Optional[datetime] = None


class Wallet(Item):
    # Properties
    name: Optional[str] = None

    # Edges
    picture: List["Photo"] = []


class Website(Item):
    # Properties
    itemType: Optional[str] = None
    url: Optional[str] = None


class WrittenWork(CreativeWork):
    pass


class Indexer(Integrator):
    # Properties
    bundleImage: Optional[str] = None
    icon: Optional[str] = None
    indexerClass: Optional[str] = None
    itemDescription: Optional[str] = None
    queryStr: Optional[str] = None
    runDestination: Optional[str] = None

    # Edges
    indexerRun: List["IndexerRun"] = []


class IndexerRun(Integrator):
    # Properties
    errorMessage: Optional[str] = None
    progress: Optional[float] = None
    progressMessage: Optional[str] = None
    queryStr: Optional[str] = None
    runStatus: Optional[str] = None
    targetDataType: Optional[str] = None

    # Edges
    indexer: List["Indexer"] = []


class CategoricalLabel(LabelAnnotation):
    # Properties
    labelValue: Optional[str] = None


class TextClassification(LabellingDataType):
    pass


class Address(Location):
    # Properties
    city: Optional[str] = None
    itemType: Optional[str] = None
    locationAutoLookupHash: Optional[str] = None
    postalCode: Optional[str] = None
    state: Optional[str] = None
    street: Optional[str] = None

    # Edges
    changelog: List["AuditItem"] = []
    country: List["Country"] = []
    location: List["Location"] = []


class Country(Location):
    # Properties
    name: Optional[str] = None

    # Edges
    flag: List["Photo"] = []
    location: List["Location"] = []


class Photo(Item):
    # Properties
    name: Optional[str] = None
    caption: Optional[str] = None
    exifData: Optional[str] = None
    height: Optional[int] = None
    width: Optional[int] = None
    channels: Optional[int] = None
    encoding: Optional[str] = None
    mode: Optional[str] = None
    ocrText: Optional[str] = None

    # Edges
    changelog: List["AuditItem"] = []
    thumbnail: List["File"] = []
    file: List["File"] = []


class Tweet(Item):
    message: Optional[str] = None
    service: Optional[str] = None
    postDate: Optional[datetime] = None
    tweetType: Optional[str] = None
    conversationId: Optional[str] = None
    retweetCount: Optional[int] = None
    replyCount: Optional[int] = None
    likeCount: Optional[int] = None

    reference: List["Tweet"] = []
    replies: List["Tweet"] = []
    mention: List[Account] = []
    author: List[Account] = []
    photo: List[Photo] = []


class Hashtag(Item):
    pass


class Message(WrittenWork):
    # Properties
    dateReceived: Optional[datetime] = None
    dateSent: Optional[datetime] = None
    externalId: Optional[str] = None
    service: Optional[str] = None
    subject: Optional[str] = None
    sourceProject: Optional[str] = None
    isMock: Optional[bool] = None

    # Edges
    message: List["Message"] = []
    messageChannel: List["MessageChannel"] = []
    photo: List["Photo"] = []
    receiver: List["Account"] = []
    sender: List["Account"] = []
    label: List["CategoricalPrediction"] = []


class Note(WrittenWork):
    # Properties
    starred: Optional[bool] = None

    # Edges
    label: List["Label"] = []


class EmailMessage(Message):
    # Properties
    starred: Optional[bool] = None

    # Edges
    bcc: List["Account"] = []
    cc: List["Account"] = []
    message: List["EmailMessage"] = []
    replyTo: List["Account"] = []


class Translation(Item):
    value: Optional[str] = None
    translatedProperty: Optional[str] = None
    srcLang: Optional[str] = None
    tgtLang: Optional[str] = None



import os
import signal
import threading
import time
from http import HTTPStatus
from threading import Thread
from typing import Callable

from loguru import logger

from pymemri.pod.api import PodError
from pymemri.pod.client import PodClient


class PluginRunStatusListener:
    def __init__(self, client, run_id, status, callback, interval=5, verbose=False):
        self.client = client
        self.run_id = run_id
        self.status = status
        self.callback = callback
        self.interval = interval
        self.verbose = verbose
        self.running = True

    def stop(self):
        if self.verbose:
            logger.info("Stopping listener...")
        self.running = False

    def run(self):
        if self.verbose:
            logger.info(f"Listening for status='{self.status}' on Item {self.run_id}")

        while self.running and threading.main_thread().is_alive():
            time.sleep(self.interval)
            try:
                run = self.client.get(self.run_id)
                if self.verbose:
                    logger.info("run status:", run.status)
                if run.status == self.status:
                    self.callback()
            except Exception as e:
                logger.error("Could not get run in status listener")


class PodHTTPStatusListener:
    def __init__(
        self,
        client: PodClient,
        run_id: str,
        callback: Callable,
        http_status: HTTPStatus,
        interval: int = 5,
        verbose: bool = False,
    ) -> None:
        """calls `self.callback` if the a pod request with pluginauth encounters self.http_status.

        Required for stopping plugins when the pod returns MISDIRECTED_REQUEST; on pod restart,
        pluginauth keys are lost, and the plugin can no longer connect to the Pod.

        Args:
            client (PodClient): the PodClient to check status for.
            run_id (str): the id of the plugin to check the http status for.
            callback (function): Function that is triggered when http_status is encountered.
            http_status (HTTPStatus): HTTPStatus received from pod that triggers callback.
            interval (int, optional): Interval the pod status is checked. Defaults to 60.
            verbose (bool, optional): Listener prints additional information if True. Defaults to False.
        """
        self.client: PodClient = client
        self.run_id = run_id
        self.callback = callback
        self.interval = interval
        self.verbose = verbose
        self.http_status = http_status
        self.running = True

    def stop(self):
        if self.verbose:
            logger.info("Stopping listener...")
        self.running = False

    def run(self):
        if self.verbose:
            logger.info(f"Listening for pod http status {self.http_status.value}")

        while self.running and threading.main_thread().is_alive():
            time.sleep(self.interval)
            try:
                _ = self.client.api.get_item(self.run_id)
                if self.verbose:
                    logger.info("run http status OK")
            except PodError as e:
                if self.verbose:
                    logger.error("run http status:", e.status)
                if e.status == self.http_status:
                    self.callback()
            except Exception as e:
                logger.error("Could not get run in httpstatus listener")


def force_exit_callback():
    logger.info("Listener aborted plugin...")
    pid = os.getpid()
    os.kill(pid, signal.SIGINT)


def get_abort_plugin_listener(client, run_id, **kwargs):
    listener = PluginRunStatusListener(
        client=client, run_id=run_id, status="aborted", callback=force_exit_callback, **kwargs
    )
    thread = Thread(
        target=listener.run,
    )
    thread.start()
    return listener


def get_pod_restart_listener(client, run_id, **kwargs):
    listener = PodHTTPStatusListener(
        client=client,
        run_id=run_id,
        callback=force_exit_callback,
        http_status=HTTPStatus.MISDIRECTED_REQUEST,
        interval=5,
        **kwargs,
    )
    thread = Thread(target=listener.run)
    thread.start()
    return listener



import sys
import uuid
from datetime import datetime, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Tuple,
    Type,
    TypeVar,
    Union,
    no_type_check,
)

from pydantic import BaseModel, Extra, PrivateAttr, validator
from pydantic.fields import Field, FieldInfo, ModelField
from pydantic.generics import GenericModel
from pydantic.main import ModelMetaclass
from typing_extensions import dataclass_transform

from .utils import get_args, get_origin, type_or_union_to_tuple, type_to_str

if TYPE_CHECKING:
    from ...pod.client import PodClient

SOURCE, TARGET, TYPE, EDGE_TYPE, LABEL, SEQUENCE, ALL_EDGES = (
    "_source",
    "_target",
    "_type",
    "_type",
    "label",
    "sequence",
    "allEdges",
)

POD_TYPES: Dict[type, str] = {
    bool: "Bool",
    str: "Text",
    int: "Integer",
    float: "Real",
    datetime: "DateTime",
}

PodType = Union[bool, str, int, float, datetime]


def _field_is_property(field: ModelField) -> bool:
    return field.outer_type_ in POD_TYPES and not field.name.startswith("_")


def _field_is_edge(field: ModelField) -> bool:
    try:
        if field.name.startswith("_") or get_origin(field.outer_type_) != list:
            return False
        args = get_args(field.outer_type_)
        if args is None or len(args) != 1:
            return False
        return True

    except Exception:
        return False


# @dataclass_transform(kw_only_default=True, field_descriptors=(Field, FieldInfo))
# class _EdgeMeta(ModelMetaclass):
#     def __call__(cls, *args: Any, **kwds: Any) -> Any:
#         """
#         Metaclass for Edge to ensure an Edge always has a generic argument (ie Edge[Item]).
#         If no generic is given to the constructor, it is inferred from type(edge.target)
#         """
#         target_annotation = cls.__annotations__["target"]
#         if isinstance(target_annotation, TypeVar):
#             # cls has no generic argument
#             generic_val = type(kwds["target"])
#             cls = cls[generic_val]  # type: ignore

#         return super().__call__(*args, **kwds)


@dataclass_transform(kw_only_default=True, field_descriptors=(Field, FieldInfo))
class _ItemMeta(ModelMetaclass):
    """
    Metaclass for ItemBase that adds required class and instance properties.

    Note: _itembase_private_attrs is a workaround for hiding private attributes
    generated by pyright/pylance signature. This is only a cosmetic difference,
    private attributres can be defined on a model with `PrivateAttr(default)` value.
    See: https://github.com/pydantic/pydantic/discussions/4563
    """

    _itembase_private_attrs = {
        "_in_pod": PrivateAttr(False),
        "__edges__": PrivateAttr(default_factory=dict),
        "_new_edges": PrivateAttr(default_factory=list),
        "_date_local_modified": PrivateAttr(default_factory=dict),
        "_original_properties": PrivateAttr(default_factory=dict),
    }

    @no_type_check  # noqa C901
    def __new__(mcs, name, bases, namespace, **kwargs) -> Any:  # noqa C901
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        cls.__property_fields__ = {k: v for k, v in cls.__fields__.items() if _field_is_property(v)}
        cls.__edge_fields__ = {k: v for k, v in cls.__fields__.items() if _field_is_edge(v)}

        # For backwards compatibility with old ItemBase
        cls.properties = list(cls.__property_fields__.keys())
        cls.edges = list(cls.__edge_fields__.keys())

        # Add private instance attributes from _itembase_private_attrs
        cls.__private_attributes__ = {**cls.__private_attributes__, **mcs._itembase_private_attrs}
        cls.__slots__ = cls.__slots__ | mcs._itembase_private_attrs.keys()
        return cls


TargetType = TypeVar("TargetType", bound="ItemBase")


class Edge(GenericModel, Generic[TargetType], smart_union=True, copy_on_model_validation=False):
    source: "ItemBase"
    target: TargetType
    name: str

    def __init__(self, source: Any = None, target: TargetType = None, name: str = None) -> None:
        super().__init__(source=source, target=target, name=name)
        # Workaround, occasionally validators will set source or target with a clone
        self.source = source
        self.target = target

    @classmethod
    def get_target_types(cls) -> Tuple[type]:
        target_annotation = cls.__annotations__["target"]
        if target_annotation == TargetType:
            return tuple()
        return type_or_union_to_tuple(target_annotation)

    @classmethod
    def get_target_types_as_str(cls) -> Tuple[str]:
        return tuple(type_to_str(t) for t in cls.get_target_types())

    @validator("target", pre=True, allow_reuse=True)
    def validate_target(cls, val: Any) -> TargetType:
        """
        To allow for overwriting schema classes on a different place, validator
        only checks by class name.

        i.e. `Edge[schema.MyItem](target=plugin.MyItem(), ...)` is allowed.
        """
        target_types = cls.get_target_types()
        if len(target_types) == 0:
            # cls has no target type annotations, validator always succeeds
            return val
        if isinstance(val, target_types):
            return val
        ttype_display = cls.__fields__["target"]._type_display()
        raise ValueError(f"target with type `{type(val).__name__}` is not a `{ttype_display}`")

    def __eq__(self, other):
        # Edge source/targets are checked by reference instead of value:
        # edge_1 == edge_2 if they have the same name, and point to the
        # same objects
        if (
            isinstance(other, Edge)
            and self.name == other.name
            and self.source is other.source
            and self.target is other.target
        ):
            return True
        return False

    def traverse(self, start: Any) -> Any:
        if start == self.source:
            return self.target
        elif start == self.target:
            return self.source
        else:
            raise ValueError("`start` is not source or target")


ItemType = TypeVar("ItemType")


class ItemBase(BaseModel, metaclass=_ItemMeta, extra=Extra.forbid):
    if TYPE_CHECKING:
        # class variables populated by the metaclass, defined here to help IDEs only
        __edge_fields__: ClassVar[Dict[str, ModelField]] = {}
        __property_fields__: ClassVar[Dict[str, ModelField]] = {}
        properties: ClassVar[List[str]] = []
        edges: ClassVar[List[str]] = []

    def __init__(self, **data: Dict[str, Any]) -> None:
        if TYPE_CHECKING:
            # instance variables populated by the metaclass, defined here to help IDEs only
            self.__edges__: Dict[str, List[Edge[Any]]] = {}
            self._in_pod: bool = True
            self._new_edges: List[str] = []
            self._date_local_modified: Dict[str, datetime] = {}
            self._original_properties: Dict[str, PodType] = {}

        # Edges are added after super().__init__
        edge_data = {}
        for edge_name in self.__edge_fields__:
            edge_data[edge_name] = data.pop(edge_name, None)

        super().__init__(**data)
        self._init_edges(**edge_data)

    def _init_edges(self, **data: Dict[str, List["ItemBase"]]) -> None:
        self.__edges__ = {k: list() for k in self.__edge_fields__.keys()}
        for edge_name, edge_targets in data.items():
            if edge_targets is not None:
                for edge_target in edge_targets:
                    self.add_edge(edge_name, edge_target)
        self._new_edges = []

    def __setattr__(self, name: str, value: Any) -> None:
        prev_val = getattr(self, name, None)
        super().__setattr__(name, value)
        if name in self.__property_fields__ and value != prev_val:
            self._date_local_modified[name] = datetime.now(timezone.utc)
            if name not in self._original_properties:
                self._original_properties[name] = prev_val

    def get_edges(self, name: str) -> List[Edge]:
        return self.__edges__[name]

    def get_all_edges(self) -> List[Edge]:
        return [edge for edge_list in self.__edges__.values() for edge in edge_list]

    def __getattribute__(self, __name: str) -> Any:
        edge_fields = super().__getattribute__("__edge_fields__")
        if __name in edge_fields:
            return [e.traverse(self) for e in super().__getattribute__("__edges__")[__name]]
        return super().__getattribute__(__name)

    def add_edge(self, edge_name: str, target: "ItemBase") -> Edge:
        """Add a new edge between `self` and `target` with name `edge_name`

        Args:
            edge_name (str): Name of the edge
            target (ItemBase): Target of the edge

        Raises:
            ValueError: Raises error if `edge_name` does not exist on this class
        """
        field = self.__edge_fields__.get(edge_name, None)
        if field is not None:
            edge = Edge[field.type_](name=edge_name, source=self, target=target)
            existing_edges = self.__edges__[field.name]
            if edge not in existing_edges:
                existing_edges.append(edge)
                self._new_edges.append(edge)
            return edge
        else:
            raise ValueError(f"{edge_name} is not an edge on {type(self).__name__}")

    @classmethod
    def pod_schema(cls) -> List[Dict[str, str]]:
        """Generates the schema as required by the Pod.

        Returns:
            List[Dict[str, str]]: List of schema definitions.
        """
        schema = []
        for field in cls.__property_fields__.values():
            property_schema = {
                "type": "ItemPropertySchema",
                "itemType": cls.__name__,
                "propertyName": field.name,
                "valueType": POD_TYPES[field.type_],
            }
            schema.append(property_schema)

        for field in cls.__edge_fields__.values():
            for target_type in type_or_union_to_tuple(field.type_):
                edge_schema = {
                    "type": "ItemEdgeSchema",
                    "edgeName": field.name,
                    "sourceType": cls.__name__,
                    "targetType": type_to_str(target_type),
                }
                schema.append(edge_schema)
        return schema

    @classmethod
    def central_schema_definition(cls) -> Dict[str, Any]:
        """Experimental feature, not covered by tests yet.

        Returns schema definition like in the central schema repository

        Returns:
            Dict[str, Any]: Central schema definition
        """
        schema = {}
        schema["description"] = getattr(cls, "description", None) or ""
        base = cls.__bases__[0]
        if base != "Item":
            schema["base_schema"] = base.__name__
        schema["properties"] = []
        for field in cls.__property_fields__.values():
            if (
                field.name not in base.properties
                or base.__property_fields__[field.name].type_ != field.type_
            ):
                property_schema = {
                    "name": field.name,
                    "type": POD_TYPES[field.type_],
                }
                schema["properties"].append(property_schema)

        schema["edges"] = []
        for field in cls.__edge_fields__.values():
            if (
                field.name not in base.edges
                or base.__edge_fields__[field.name].type_ != field.type_
            ):
                target_types = type_or_union_to_tuple(field.type_)
                ttype_json = [type_to_str(t) for t in target_types]
                if len(ttype_json) == 1:
                    ttype_json = ttype_json[0]
                    edge_schema = {"name": field.name, "target": ttype_json}
                    schema["edges"].append(edge_schema)

        return schema

    @property
    def _updated_properties(self):
        return set(self._original_properties.keys())

    def reset_local_sync_state(self):
        """
        reset_local_sync_state is called when self is created or updated (optionally via bulk) in the PodClient.
        """
        self._original_properties = dict()
        self._date_local_modified = dict()
        self._in_pod = True

    @staticmethod
    def create_id() -> str:
        return uuid.uuid4().hex

    def create_id_if_not_exists(self):
        if hasattr(self, "id") and getattr(self, "id") is None:
            self.id = self.create_id()

    def store(self, client: "PodClient") -> None:
        return client.add_to_store(self)

    def __str__(self) -> str:
        return f"{type(self).__name__}(#{self.id})"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.dict() == other.dict()

    def _iter(
        self,
        *,
        to_dict: bool = False,
        include: set = None,
        exclude: set = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        if exclude is None:
            exclude = set(type(self)._itembase_private_attrs)
        else:
            exclude = exclude + set(type(self)._itembase_private_attrs)
        return super()._iter(
            to_dict=to_dict,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    @staticmethod
    def _datetime_to_timestamp(dt: datetime) -> int:
        """Convert datetime to Pod timestamp format"""
        return round(dt.timestamp() * 1000)

    def property_dict(
        self,
        exclude_none: bool = True,
        datetime_to_timestamp: bool = True,
    ) -> Dict[str, PodType]:
        """Returns a dict of values of all properties of self

        Args:
            exclude (bool, optional): Remove `None` values. Defaults to True.
            datetime_to_timestamp (bool, optional): convert datetime values to timestamp in milliseconds. Defaults to True.

        Returns:
            Dict[str, PodType]: Dictionary of all property values
        """
        include = self.__property_fields__.keys()
        res = self.dict(include=include, exclude_none=exclude_none)
        for k, v in res.items():
            if datetime_to_timestamp and isinstance(v, datetime):
                res[k] = self._datetime_to_timestamp(v)
        return res

    def to_json(self) -> Dict[str, PodType]:
        """Returns dict of property values along with the type of self.
        This is the format as expected by the Pod.

        Returns:
            Dict[str, PodType]: Dictionary of all property values
        """
        res = self.property_dict()
        res["type"] = type(self).__name__
        return res

    @classmethod
    def from_data(cls: Type[ItemType], **kwargs: Dict[str, Any]) -> ItemType:
        # TODO remove and use __init__
        return cls(**kwargs)

    @classmethod
    def from_json(
        cls: Type[ItemType], json: Dict[str, Any], properties_only: bool = True
    ) -> ItemType:
        if properties_only:
            json = {k: v for k, v in json.items() if k in cls.properties}
        return cls(**json)


Edge.update_forward_refs()
ItemBase.update_forward_refs()



from pathlib import Path

from loguru import logger

from ..gitlab_api import DEFAULT_PACKAGE_VERSION, GitlabAPI, find_git_repo

DEFAULT_PLUGIN_MODEL_PACKAGE_NAME = "plugin-model-package"
DEFAULT_PYTORCH_MODEL_NAME = "pytorch_model.bin"
DEFAULT_HUGGINFACE_CONFIG_NAME = "config.json"


def write_huggingface_model_to_package_registry(
    project_name, model, version=DEFAULT_PACKAGE_VERSION, client=None
):
    import torch

    api = GitlabAPI(client=client)
    project_id = api.project_id_from_name(project_name)
    local_save_dir = Path("/tmp")
    torch.save(model.state_dict(), local_save_dir / DEFAULT_PYTORCH_MODEL_NAME)
    model.config.to_json_file(local_save_dir / DEFAULT_HUGGINFACE_CONFIG_NAME)

    for f in [DEFAULT_HUGGINFACE_CONFIG_NAME, DEFAULT_PYTORCH_MODEL_NAME]:
        file_path = local_save_dir / f
        logger.info(
            f"writing {f} to package registry of {project_name} with project id {project_id}"
        )
        api.write_file_to_package_registry(
            project_id,
            file_path,
            package_name=DEFAULT_PLUGIN_MODEL_PACKAGE_NAME,
            version=version,
            trigger_pipeline=False,
        )
    api.trigger_pipeline(project_id=project_id)


def write_model_to_package_registry(model, project_name=None, client=None):
    project_name = project_name if project_name is not None else find_git_repo()
    if type(model).__module__.startswith("transformers"):
        import torch
        import transformers
    if isinstance(model, transformers.PreTrainedModel):
        write_huggingface_model_to_package_registry(project_name, model, client=client)
    else:
        raise ValueError(f"Model type not supported: {type(model)}")


def download_huggingface_model_for_project(
    project_path=None, files=None, download_if_exists=False, client=None
):
    api = GitlabAPI(client=client)
    if files is None:
        files = ["config.json", "pytorch_model.bin"]
    for f in files:
        out_file_path = api.download_package_file(
            f, project_path=project_path, package_name=DEFAULT_PLUGIN_MODEL_PACKAGE_NAME
        )
    return out_file_path.parent


def load_huggingface_model_for_project(
    project_path=None, files=None, download_if_exists=False, client=None
):
    out_dir = download_huggingface_model_for_project(
        project_path, files, download_if_exists, client=client
    )
    from transformers import AutoModelForSequenceClassification

    model = AutoModelForSequenceClassification.from_pretrained(out_dir)
    return model



from pathlib import Path
from typing import Any, ClassVar, List, Optional, Union

from pydantic import PrivateAttr

from ...exporters.exporters import Query
from ._central_schema import Dataset


def filter_rows(dataset: dict, filter_val=None) -> dict:
    missing_idx = set()
    for column in dataset.values():
        missing_idx.update([i for i, val in enumerate(column) if val == filter_val])
    return {
        k: [item for i, item in enumerate(v) if i not in missing_idx] for k, v in dataset.items()
    }


class Dataset(Dataset):
    requires_client_ref: ClassVar[bool] = True
    _client: Optional[Any] = PrivateAttr(None)

    def _get_items(self):
        if self._client is None:
            raise ValueError("Dataset does not have associated PodClient.")
        if not len(self.entry):
            for e in self._client.get_edges(self.id):
                self.add_edge(e["name"], e["item"])

        return self.entry

    def _get_data(self, dtype: str, columns: List[str], filter_missing: bool = True):
        if self._client is None:
            raise ValueError("Dataset does not have associated PodClient.")
        items = self._get_items()

        query = Query("id", *columns)
        result = query.execute(self._client, items)
        if filter_missing:
            result = filter_rows(result, filter_val=None)
        return query.convert_dtype(result, dtype)

    def _infer_columns(self):
        column_names = []
        for feature in self.feature:
            column_name = "data." + feature.propertyName
            column_names.append(column_name)

        # TODO infer columns for different label types
        column_names.append("annotation.labelValue")

        return column_names

    def to(self, dtype: str, columns: Optional[List[str]] = None, filter_missing: bool = True):
        """
        Converts Dataset to a different format.

        Available formats:
        list: a 2-dimensional list, containing one dataset entry per row
        dict: a list of dicts, where each dict contains {column: value} for each column
        pd: a Pandas dataframe


        Args:
            dtype (str): Datatype of the returned dataset
            columns (List[str]): Column names of the dataset
            filter_missing (bool, optional): If true, all rows that contain `None` values are omitted.
                Defaults to True.

        Returns:
            Any: Dataset formatted according to `dtype`
        """
        if columns is None:
            columns = self._infer_columns()
        return self._get_data(dtype, columns, filter_missing)

    def save(self, path: Union[Path, str], columns: List[str], filter_missing: bool = True):
        """
        Save dataset to CSV.
        """
        result = self._get_data("pandas", columns, filter_missing)
        result.to_csv(path, index=False)
